import logging
log = logging.getLogger(__name__)

import coralme

#def add_ribosome(me_model, ribosome_stoich, ribosome_subreactions, rrna_mods, modification_info, verbose = True):
def add_ribosome(me_model, ribosome_stoich, ribosome_subreactions, rrna_mods, verbose = True):
	lst = ['generic_Era', 'generic_RbfA', 'generic_RimM']
	for ComplexData in lst:
		if len(me_model.process_data.query(ComplexData)) == 0:
			logging.warning('ComplexData \'{:s}\' not in the ME-model. The ME-model would be unfeasible.'.format(ComplexData))
			logging.warning('Check if the behavior is correct or if \'{:s}\' is present in the \'Generic Complex ID\' column of the organism-specific matrix.'.format(ComplexData.replace('generic_', '')))

	ribosome_complex = coralme.core.processdata.ComplexData(me_model.global_info['ribosome_id'], me_model)

	for idx, mod_data in rrna_mods.iterrows():
		for position in mod_data.positions.split(','):
			rrna_mod = coralme.core.processdata.SubreactionData('{:s}_at_{:s}'.format(mod_data.modification, position), me_model)
			rrna_mod.enzyme = mod_data.enzymes.split(' AND ') if mod_data.enzymes != 'No_Machine' else ['CPLX_dummy']
			#rrna_mod.stoichiometry = modification_info[mod_data.modification]['metabolites']
			rrna_mod.stoichiometry = me_model.process_data.get_by_id(mod_data.modification).stoichiometry
			rrna_mod.keff = 65. # iOL uses 65. for all RNA mods

			# Add element contribution from modification to rRNA
			#rrna_mod._element_contribution = modification_info[mod_data.modification]['elements']
			rrna_mod._element_contribution = me_model.process_data.get_by_id(mod_data.modification).calculate_element_contribution()

			#if 'carriers' in modification_info[mod_data.modification].keys():
				#for carrier, stoich in modification_info[mod_data.modification]['carriers'].items():
					#if stoich < 0:
						#rrna_mod.enzyme += [carrier]
					#rrna_mod.stoichiometry[carrier] = stoich
			ribosome_complex.subreactions[rrna_mod.id] = 1.

	for subreaction_id in ribosome_subreactions:
		if not me_model.process_data.has_id('gtp_hydrolysis_era'):
			stoichiometry = { 'gtp_c': -2.0, 'h2o_c': -2.0, 'gdp_c': 2.0, 'h_c': 2.0, 'pi_c': 2.0 }
			coralme.util.building.add_subreaction_data(
				me_model, modification_id = 'gtp_hydrolysis_era', modification_stoichiometry = stoichiometry, modification_enzyme = None)

		# add subreaction to model
		if me_model.process_data.has_id(subreaction_id):
			subreaction = me_model.process_data.get_by_id(subreaction_id)
		else:
			subreaction = coralme.core.processdata.SubreactionData(subreaction_id, me_model)
			#subreaction.stoichiometry = ribosome_subreactions[subreaction_id]['stoich']
			#reaction_id = me_model.global_info['translation_subreactions'][subreaction_id]
			reaction_id = me_model.global_info['translation_subreactions'].get(subreaction_id,None)
			if reaction_id == '' or reaction_id is None:
				# If reaction_id is not in global_info it must have been defined in subreaction_matrix
				subreaction.stoichiometry = {}
			else:
				subreaction.stoichiometry = me_model.process_data.get_by_id(reaction_id).stoichiometry

		subreaction.enzyme = ribosome_subreactions[subreaction_id]['enzymes']
		# account for subreactions in complex data. num_mods is always 1
		ribosome_complex.subreactions[subreaction.id] = 1.

	# Ribosomes in iOL1650 contain 171 mg2 ions
	ribosome_complex.subreactions['mod_mg2_c'] = me_model.global_info['mg2_per_ribosome']
	ribosome_components = ribosome_complex.stoichiometry
	for process in ribosome_stoich:
		for protein, amount in ribosome_stoich[process]['stoich'].items():
			ribosome_components[protein] += amount

	ribosome_complex.create_complex_formation(verbose = verbose)



import tqdm
bar_format = '{desc:<75}: {percentage:.1f}%|{bar:10}| {n_fmt:>5}/{total_fmt:>5} [{elapsed}<{remaining}]'
import coralme
import logging

#def add_trna_modification_procedures(me_model, trna_mods, modification_info):
def add_trna_modification_procedures(me_model, trna_mods):
	# remove duplications
	trna_mods = trna_mods.drop_duplicates(['modification', 'positions'], keep = 'first')
	for idx, mod_data in tqdm.tqdm(list(trna_mods.iterrows()), 'Adding tRNA modification SubReactions...', bar_format = bar_format):
		for position in mod_data['positions'].split(','):
			#if mod_data.type == 'met_tRNA':
			#if mod_data['bnum'] in me_model.global_info['START_tRNA']:
				##name = '{:s}_at_{:s}_in_{:s}'.format(mod_data.modification, position, mod_data.type)
				#name = '{:s}_at_{:s}_in_met_tRNA'.format(mod_data['modification'], position)
			#else:
			name = '{:s}_at_{:s}'.format(mod_data['modification'], position)
			trna_mod = coralme.core.processdata.SubreactionData(name, me_model)
			#trna_mod.enzyme = mod_data['enzymes'].split(' AND ') if mod_data['enzymes'] != 'No_Machine' else None
			trna_mod.enzyme = mod_data['enzymes'].split(' AND ') if mod_data['enzymes'] != 'No_Machine' else ['CPLX_dummy']
			#trna_mod.stoichiometry = modification_info[mod_data.modification]['metabolites']
			if me_model.process_data.has_id(mod_data['modification']):
				trna_mod.stoichiometry = me_model.process_data.get_by_id(mod_data['modification']).stoichiometry
			else:
				logging.warning('The tRNA modification \'{:s}\' was not added into the ME-model. Please, add it in the subreaction.txt input file.'.format(mod_data['modification']))
				trna_mod.stoichiometry = {}

			trna_mod.keff = 65.  # iOL uses 65 for all tRNA mods

			for met, stoich in trna_mod.stoichiometry.items():
				if not me_model.metabolites.has_id(met):
					logging.warning("Creating metabolite {} in {}".format(met,trna_mod.id))
					met_obj = coralme.core.component.Metabolite(met)
					me_model.add_metabolites([met_obj])
				if isinstance(me_model.metabolites.get_by_id(met), coralme.core.component.Complex) and stoich < 0:
					trna_mod.enzyme.append(met)

		# Add element contribution from modification to tRNA
		#trna_mod._element_contribution = modification_info[mod_data.modification]['elements']
		if me_model.process_data.has_id(mod_data['modification']):
			trna_mod._element_contribution = me_model.process_data.get_by_id(mod_data['modification']).calculate_element_contribution()
		else:
			trna_mod._element_contribution = {}



import cobra
import coralme
from collections import Counter

# Extend Counter to let us multiply by a int or float
# https://stackoverflow.com/questions/23677982/python-how-to-multiply-values-of-a-counter-object
# Original code modified to allow multiplication by a float
class MCounter(Counter):
	"""This is a slight extension of the ``Collections.Counter`` class
	to also allow multiplication with integers."""

	def __mul__(self, other):
		if not isinstance(other, (int, float)):
			raise TypeError("Non-int factor")
		return MCounter({k: other * v for k, v in self.items()})

	def __rmul__(self, other):
		return self * other  # call __mul__

	def __add__(self, other):
		return MCounter(super().__add__(other))

class MEComponent(cobra.core.metabolite.Metabolite):
	"""
	coralme component representation. Inherits from
	:class:`cobra.core.metabolite.Metabolite`

	Parameters
	----------
	id : str
		Identifier of the component. Should follow best practices of child
		classes

	"""
	def __init__(self, id):
		cobra.core.metabolite.Metabolite.__init__(self, id)

	def get_fluxes(self,
		only_types = (), ignore_types = (),
		threshold = 0., flux_dict = 0, include_zeroes = True,
		solution = None, keffs = False, verbose = False):
		"""
		Shortcut for coralme.util.flux_analysis.flux_based_reactions(model, met_id, [...])
		"""
		return coralme.util.flux_analysis.flux_based_reactions(self._model, self.id, only_types, ignore_types, threshold, flux_dict, include_zeroes, solution, keffs, verbose)

	def add_boundary(self, type: str = "exchange", reaction_id = None, lb = None, ub = None, sbo_term = None):
		"""
		Shortcut for model.add_boundary(model.metabolites.get_by_id(met_id), [...])
		"""
		if self._model is not None:
			self._model.add_boundary(self, type, reaction_id, lb, ub, sbo_term)

	def remove_from_me_model(self, method = 'subtractive'):
		"""
		Remove metabolite from ME-model along with any relevant
		:class:`coralme.core.processdata.ProcessData`

		Parameters
		----------
		method : str
			- destructive: remove metabolite from model and remove reactions
			  it is involved in.
			- subtractive: remove only metabolite from model.

		"""
		if self.id in self._model.process_data:
			self._model.process_data.remove(self.id)

		# If cannot import SymbolicParameter, assume using cobrapy versions <= 0.5.11
		#try: from optlang.interface import SymbolicParameter
		#except ImportError: self.remove_from_model(method = method)
		#else:
		if method.lower() == 'subtractive':
			self.remove_from_model(destructive = False)
		elif method.lower() == 'destructive':
			self.remove_from_model(destructive = True)
		else:
			raise AttributeError("method must be \'subtractive\' or \'destructive\'")

	@property
	def complexes(self):
		return coralme.builder.helper_functions.find_complexes(self,seen= set())
	@property
	def functions(self):
		cplxs = self.complexes
		functions = set()
		for c in cplxs:
			functions = functions | coralme.builder.helper_functions.get_functions(c)
		return functions

	def molecular_weight(self, numeric = True):
		res = self.formula_weight
		if numeric:
			return res
		else:
			return '{:f} g/mol'.format(res)

class Metabolite(MEComponent):
	"""
	coralme metabolite representation

	Parameters
	----------
	id : str
		Identifier of the metabolite

	"""
	def __init__(self, id):
		MEComponent.__init__(self, id)

	#def __repr__(self):
		#return 'Metabolite'

class TranscribedGene(MEComponent):
	"""
	Metabolite class for gene created from
	:class:`coralme.core.reaction.TranscriptionReaction`

	Parameters
	----------
	id : str
		Identifier of the transcribed gene. As a best practice, this ID should
		be prefixed with 'RNA + _'

	RNA_type : str
		Type of RNA encoded by gene sequence (mRNA, rRNA, tRNA, or ncRNA)

	nucleotide_sequence : str
		String of base pair abbreviations for nucleotides contained in the gene

	Attributes
	----------
	left_pos : int
		Left position of gene on the sequence of the (+) strain

	right_pos : int
		Right position of gene on the sequence of the (+) strain

	strand : str
		- (+) if the RNA product is on the leading strand
		- (-) if the RNA product is on the comple(mentary strand

	"""

	def __init__(self, id, rna_type, nucleotide_sequence):
		MEComponent.__init__(self, id)
		self.left_pos = None
		self.right_pos = None
		self.strand = None
		self.RNA_type = rna_type
		self.nucleotide_sequence = nucleotide_sequence

	#def __repr__(self):
		#return 'TranscribedGene'

	@property
	def codon_usage(self):
		if self.RNA_type == 'mRNA':
			seq = self.nucleotide_sequence
			codons = [ str(seq[pos:pos+3]) for pos in range(0, len(seq), 3) ]
			return Counter(codons)
		else:
			return None

	@property
	def nucleotide_count(self):
		"""
		Get number of each nucleotide monophosphate

		Returns
		-------
		dict
			{nucleotide_monophosphate_id: count}

		"""
		seq = self.nucleotide_sequence
		counts = {i: seq.count(i) for i in ("A", "T", "G", "C")}
		#monophosphate_counts = { coralme.util.dogma.transcription_table[k].replace("tp_c", "mp_c"): v for k, v in counts.items() }
		monophosphate_counts = { coralme.util.dogma.transcription_table['c'][k].replace("tp_c", "mp_c"):v for k,v in counts.items() }
		return monophosphate_counts

class TranslatedGene(MEComponent):
	"""
	Metabolite class for protein created from
	:class:`coralme.core.reaction.TranslationReaction`

	Parameters
	----------
	id : str
		Identifier of the translated protein product. Should be prefixed
		with "protein + _"

	"""
	def __init__(self, id):
		MEComponent.__init__(self, id)

	#def __repr__(self):
		#return 'TranslatedGene'

	@property
	def translation_data(self):
		"""
		Get translation data that defines protein.

		Assumes that TranslatedGene is "protein + _ + <translation data id>"

		Returns
		-------
		:class:`coralme.core.processdata.TranslationData`
			Translation data used to form translation reaction of protein
		"""
		locus = self.id.replace('protein_', '')
		return self._model.process_data.get_by_id(locus)

# 	@property
# 	def complexes(self):
# 		"""Get the complexes that the protein forms

# 		Returns
# 		-------
# 		list
# 			List of :class:`coralme.core.component.Complex` s that the protein
# 			is a subunit of
# 		"""
# 		complex_list = []
# 		for reaction in self.reactions:
# 			if hasattr(reaction, 'complex'):
# 				complex_list.append(reaction.complex)
# 		return complex_list

	@property
	def metabolic_reactions(self):
		"""Get the mtabolic reactions that the protein helps catalyze

		Returns
		-------
		list
			List of :class:`coralme.core.reactions.MetabolicReaction` s
			that the protein helps catalyze
		"""
		metabolic_reactions = []
		for complexes in self.complexes:
			metabolic_reactions.extend(complexes.metabolic_reactions)
		return metabolic_reactions

	@property
	def amino_acid_sequence(self):
		"""
		Get amino acid sequence of protein

		Returns
		-------
		str
			Amino acid sequence of protein

		"""
		return self.translation_data.amino_acid_sequence

class ProcessedProtein(MEComponent):
	"""
	Metabolite class for protein created from
	:class:`coralme.core.reaction.PostTranslationReaction`

	Parameters
	----------
	id : str
		Identifier of the processed protein

	unprocessed_protein_id : str
		Identifier of protein before being processed by PostTranslationReaction

	"""
	def __init__(self, id, unprocessed_protein_id):
		MEComponent.__init__(self, id)
		self.unprocessed_protein_id = unprocessed_protein_id

	#def __repr__(self):
		#return 'ProcessedProtein'

	@property
	def unprocessed_protein(self):
		"""Get unprocessed protein reactant in PostTranslationReaction

		Returns
		-------
		:class:`coralme.core.component.TranslatedGene`
			Unprocessed protein object
		"""
		return self._model.metabolites.get_by_id(self.unprocessed_protein_id)

class Complex(MEComponent):
	"""
	Metabolite class for protein complexes

	Parameters
	----------
	id : str
		Identifier of the protein complex.
	"""

	def __init__(self, id):
		MEComponent.__init__(self, id)

	#def __repr__(self):
		#return 'Complex'

	@property
	def metabolic_reactions(self):
		"""Get metabolic reactions catalyzed by complex

		Returns
		-------
		list
			List of :class:`coralme.core.reaction.MetabolicReaction` s
			catalyzed by complex.
		"""
		reaction_list = []
		for reaction in self.reactions:
			if hasattr(reaction, 'stoichiometric_data'):
				reaction_list.append(reaction)
		return reaction_list

class Ribosome(Complex):
	"""
	Metabolite class for Ribosome complexes. Inherits from
	:class:`coralme.core.component.Complex`

	Parameters
	----------
	id : str
		Identifier of the Ribosome.
	"""
	def __init__(self, id):
		Complex.__init__(self, id)

	#def __repr__(self):
		#return 'Ribosome'

class RNAP(Complex):
	"""
	Metabolite class for RNA polymerase complexes. Inherits from
	:class:`coralme.core.component.Complex`

	Parameters
	----------
	id : str
		Identifier of the RNA Polymerase.
	"""

	def __init__(self, id):
		Complex.__init__(self, id)

	#def __repr__(self):
		#return 'RNAP'

class GenericComponent(MEComponent):
	"""
	Metabolite class for generic components created from
	:class:`coralme.core.reaction.GenericFormationReaction`

	Parameters
	----------
	id : str
		Identifier of the generic tRNA. As a best practice should follow
		template: 'generic + _ +  <generic metabolite id>'
	"""
	def __init__(self, id):
		MEComponent.__init__(self, id)

	#def __repr__(self):
		#return 'GenericComponent'

class GenerictRNA(MEComponent):
	"""
	Metabolite class for generic tRNAs created from
	:class:`coralme.core.reaction.tRNAChargingReaction`

	Parameters
	----------
	id : str
		Identifier of the generic tRNA. As a best practice should follow
		template: 'generic_tRNA + _ + <codon> + _ + <amino acid metabolite id>'

	"""
	def __init__(self, id):
		MEComponent.__init__(self, id)

	#def __repr__(self):
		#return 'GenerictRNA'

class Constraint(MEComponent):
	"""
	Metabolite class for global constraints such as biomass

	Parameters
	----------
	id : str
		Identifier of the constraint
	"""
	def __init__(self, id):
		MEComponent.__init__(self, id)

	#def __repr__(self):
		#return 'Constraint'

def create_component(component_id, default_type = MEComponent, rnap_set = set()):
	"""creates a component and attempts to set the correct type"""
	if not isinstance(component_id, str):
		raise TypeError("Component ID \'{:s}\' must be a str, not \'{:s}\'.".format(repr(component_id), str(type(component_id))))
	if component_id.startswith("protein_"):
		return TranslatedGene(component_id)
	elif component_id.startswith("RNA_"):
		raise ValueError(
			'TranscribedGene \'{:s}\' should not be added using '
			'create_component. It requires additional information '
			'when creating instance.'.format(component_id)
			)
	elif component_id.startswith("ribosome"):
		return Ribosome(component_id)
	elif component_id.startswith("RNA_Polymerase") or component_id in rnap_set:
		return RNAP(component_id)
	elif component_id.startswith("generic_tRNA"):
		return GenerictRNA(component_id)
	elif component_id.endswith('_c'):
		return Metabolite(component_id)
	elif component_id.startswith('generic_'):
		return GenericComponent(component_id)
	else:
		return default_type(component_id)



import logging
import coralme

mmol = 6.022e20  # number of molecules per mmol
nm2_per_m2 = 1e18  # used to convert nm^2 to m^2

# add_translocation_pathways helper function
def add_translocation_data_and_reaction(model, pathways, preprocessed_id, processed_id, compartment, peptide_data, multipliers, membrane_constraints, alt = False):
	suffix = '_alt' if alt else ''

	data_id = 'translocation_' + preprocessed_id + '_' + compartment + suffix
	if not model.process_data.has_id(data_id):
		data = coralme.core.processdata.PostTranslationData(data_id, model, processed_id, preprocessed_id)
		data.translocation = pathways
		#data.translocation_multipliers = multipliers_protein_keys.get(preprocessed_id, {})
		data.translocation_multipliers = multipliers.get(preprocessed_id, {})

		# Add protein surface area constraint
		if membrane_constraints and compartment != 'Periplasm':
			protein = peptide_data.protein
			protein_met = model.metabolites.get_by_id('protein_' + protein)
			mass = protein_met.formula_weight / 1000.  # in kDa
			membrane_thickness = model.global_info['membrane_thickness']
			thickness = membrane_thickness[compartment]
			# Relationship uses protein molecular in kDa
			# Adds surface area constraint in units of m^2/mmol
			data.surface_area['SA_protein_' + compartment] = (1.21 / thickness * 2.) * mass * mmol / nm2_per_m2

		rxn = coralme.core.reaction.PostTranslationReaction('translocation_' + peptide_data.id + '_' + compartment + suffix)
		rxn.posttranslation_data = data
		model.add_reactions([rxn])
		rxn.update()

def add_translocation_pathways(model, pathways_df, abbreviation_to_pathway, multipliers, membrane_constraints = False):
	# loop through all translation data and add translocation rxns/surface area constraints if they are membrane proteins
	# We can save time here if filtering what we need to process, not iterate over all TranslationData
	#for peptide_data in model.translation_data:
	for idx, translocation_info in pathways_df.iterrows():
		if not model.process_data.has_id(translocation_info['Protein']):
			continue

		peptide_data = model.process_data.get_by_id(translocation_info['Protein'])
		# extract translocation info if peptide contained in complex stoichiometry
		#translocation_info = pathways_df[pathways_df['Protein'].str.match(peptide_data.id)]
		# iterate if protein is not in a membrane complex
		#if len(translocation_info) == 0:
			#continue

		# Assign preprocessed and processed (translocated) peptide ids
		compartment = translocation_info['Protein_compartment'] #.values[0]
		processed_id = 'protein_' + peptide_data.id + '_' + compartment
		preprocessed_id = 'protein_' + peptide_data.id

		# compile translocation pathways for each membrane protein
		pathways = set()
		pathways_alt = set()
		for abbrev in translocation_info['translocase_pathway']: #.values[0]:
			try:
				pathway_name = abbreviation_to_pathway[abbrev]
			except:
				continue

			# The tat translocation pathway can use an alternate enzyme
			if isinstance(pathway_name, list):
				pathways.add(pathway_name[0])
				pathways_alt.add(pathway_name[1])
			else:
				pathways.add(pathway_name)
				pathways_alt.add(pathway_name)

		# Call the helper function
		add_translocation_data_and_reaction(
			model, pathways, preprocessed_id, processed_id, compartment, peptide_data, multipliers, membrane_constraints, alt = False)
		# if there's an alternative pathway (tat) add this reaction as well
		if pathways != pathways_alt:
			add_translocation_data_and_reaction(
				model, pathways_alt, preprocessed_id, processed_id, compartment, peptide_data, multipliers, membrane_constraints, alt = True)

def add_lipoprotein_formation(model, compartment_dict, lipoprotein_precursors, lipid_modifications, membrane_constraints = False, update = True):
	# add_lipoprotein_formation helper function
	def add_lipoprotein_data_and_reaction(first_lipid, second_lipid):

		# Add PostTranslation Data, modifications and surface area
		data = coralme.core.processdata.PostTranslationData(reaction_prefix + '_' + second_lipid, model, processed_id, preprocessed_id)
		data.subreactions['mod_1st_' + first_lipid] = 1
		data.subreactions['mod_2nd_' + second_lipid] = 1
		data.biomass_type = 'lipid_biomass'

		if membrane_constraints:
			thickness_dict = model.global_info['membrane_thickness']
			thickness = thickness_dict['Outer_Membrane']

			# From Liu et al. x2 for each to account for each leaflet
			protein_SA = 1.21 / thickness * 2 * mass * mmol / nm2_per_m2
			data.surface_area = {
				'SA_protein_' + compartment: -protein_SA,
				'SA_lipoprotein': 1. * mmol / nm2_per_m2
				}

		# Add Reaction to model and associated it with its data
		rxn = coralme.core.reaction.PostTranslationReaction(reaction_prefix + '_' + second_lipid)
		model.add_reactions([rxn])
		rxn.posttranslation_data = data

		if update:
			rxn.update()

	# loop through all proteins which need lipid modifications (lipoproteins)
	for protein in lipoprotein_precursors.values():
		compartment = compartment_dict.get(protein)
		if model.metabolites.has_id('protein_' + protein) and compartment is not None:
			protein_met = model.metabolites.get_by_id('protein_' + protein)
			mass = protein_met.formula_weight / 1000.  # in kDa

			processed_id = 'protein_' + protein + '_lipoprotein_' + compartment
			preprocessed_id = 'protein_' + protein + '_' + compartment

			for mod in lipid_modifications:
				reaction_prefix = protein + '_lipid_modification_' + mod
				add_lipoprotein_data_and_reaction(mod, 'pe160_p')
				add_lipoprotein_data_and_reaction(mod, 'pg160_p')

		if model.metabolites.has_id('protein_' + protein) and compartment is None:
			logging.warning('A lipoprecursor protein \'{:s}\' has not compartment. Please check the list of IDs in \'lipoprotein_precursors\' and the compartment of the protein in the OSM.'.format(protein))

		if not model.metabolites.has_id('protein_' + protein):
			logging.warning('A lipoprecursor protein \'{:s}\' is not in the ME-model. Please check if the ID in \'lipoprotein_precursors\' are in the OSM or if the gene is a pseudogene.'.format(protein))

		else:
			continue
			#print('problem with', 'protein_' + protein)



import copy
import cobra
import coralme

# Written originally by Rodrigo Santibanez for ME model
def perform_gene_knockouts(model, genes, mets_to_test = []):
	if isinstance(genes, (str, coralme.core.component.TranscribedGene)):
		genes = set([genes])

	if isinstance(model, coralme.core.model.MEModel):
		test = model.copy()
		for gene in genes:
			gene = gene.id if isinstance(gene, coralme.core.component.TranscribedGene) else gene
			gene = 'RNA_{:s}'.format(gene) if not gene.startswith('RNA_') else gene # only valid for ME-models
			if model.metabolites.has_id(gene):
				for TU in test.transcription_data:
					data = test.transcription_data.get_by_id(TU.id)
					test.transcription_data.get_by_id(TU.id).RNA_products = data.RNA_products.difference([gene])
				for rxn in test.reactions.query('transcription_'):
					rxn.update()
			else:
				raise AttributeError('Gene ID \'{:s}\' is not in the model.'.format(gene))
	else:
		test = copy.deepcopy(model)
		for gene in genes:
			gene = gene.id if isinstance(gene, cobra.core.gene.Gene) else gene
			# similar to cobra.manipulation.delete.remove_genes(model, genes, remove_reactions = True)
			if test.genes.has_id(gene):
				test.genes.get_by_id(gene).knock_out()
			else:
				raise AttributeError('Gene ID \'{:s}\' is not in the model.'.format(gene))

	for met in mets_to_test:
		if test.metabolites.has_id(met) and not test.reactions.has_id('SK_{:s}'.format(met)):
			test.add_boundary(test.metabolites.get_by_id(met), type = 'sink', lb = 0., ub = 0.)
		else:
			raise AttributeError('Metabolite ID \'{:s}\' is not in the model or a sink reaction already exists.'.format(gene))

	return test

def check_knockout_using_qminos(model, genes, optTol = 1e-15, feasTol = 1e-15):
	# test = copy.deepcopy(model)
	test = perform_gene_knockouts(model, genes)

	nlp = coralme.core.model.MEModel.construct_lp_problem(test, lambdify = True, as_dict = True, per_position = False)
	solver = coralme.solver.solver.ME_NLP(**nlp)
	solver.opt_realdict['lp']['Optimality tol'] = optTol
	solver.opt_realdict['lp']['Feasibility tol'] = feasTol

	if isinstance(test, coralme.core.model.MEModel):
		muopt, xopt, yopt, zopt, basis, stat = solver.bisectmu()
	else:
		xopt, yopt, zopt, stat, basis = solver.solvelp(1., None, 'double')
		muopt = float(sum([ x*c for x,c in zip(xopt, nlp['c']) if c != 0 ]))

	solution = coralme.core.model.MEModel._solver_solution_to_cobrapy_solution(test, muopt, xopt, yopt, zopt, stat)
	return solution

def create_ko_model_in_lp_format(model, genes, growth_rate, mets_to_test, *args):
	if len(args) == 4:
		model, genes, growth_rate, mets_to_test = args

	test = perform_gene_knockouts(model, genes, mets_to_test)
	nlp = test.construct_lp_problem(lambdify = False, as_dict = True, per_position = True)
	nlp['Sf'], nlp['Se'], nlp['xl'], nlp['xu'] = coralme.builder.helper_functions.evaluate_lp_problem(nlp['Sf'], nlp['Se'], nlp['xl'], nlp['xu'], { test.mu : growth_rate }, nlp['mu'])

	indexes = { met:(test.reactions._dict['SK_{:s}'.format(met)], test.metabolites._dict[met]) for met in mets_to_test }

	return nlp, indexes

def check_many_mets_at_a_time(args):
	for mid, (rxn, met) in args[1].items():
		# print(args[0]['xl'][rxn], args[0]['xu'][rxn])
		args[0]['xl'][rxn] = -1000.
		args[0]['xu'][rxn] = +1000.
		# print(args[0]['xl'][rxn], args[0]['xu'][rxn])

	nlp = args[0]
	xopt, yopt, zopt, stat, basis = coralme.solver.solver.ME_NLP(**nlp).solvelp(muf = None, basis = None, precision = 'quad')
	muopt = [ x*c for x,c in zip(xopt, nlp['c']) if c != 0 ][0]
	sol = coralme.core.model.MEModel._solver_solution_to_cobrapy_solution((nlp['Lr'], nlp['Lm']), muopt, xopt, yopt, zopt, stat)
	return sol

def check_all_mets_at_a_time(nlp, indexes):
	return check_many_mets_at_a_time((nlp, indexes))

def get_reduced_costs_from_nlp(nlp, objective_value = 0.1):
	xopt, yopt, zopt, stat, basis = coralme.solver.solver.ME_NLP(**nlp).solvelp(muf = objective_value, basis = None, precision = 'quad')
	muopt = [ x*c for x,c in zip(xopt, nlp['c']) if c != 0 ][0]
	sol = coralme.core.model.MEModel._solver_solution_to_cobrapy_solution((nlp['Lr'], nlp['Lm']), muopt, xopt, yopt, zopt, stat)
	return sol.reduced_costs

def get_reduced_costs_from_model(model, objective_value = 0.1, target_reaction = 'biomass_dilution'):
	if not model.reactions.has_id(target_reaction):
		raise AttributeError('Model has no reaction \'{:s}\''.format(target_reaction))

	nlp = model.construct_lp_problem(lambdify = False, as_dict = True, per_position = True)

	# change objective function and its bounds
	rxn_id = { x:idx for idx,x in enumerate(nlp['Lr']) }

	# remove objective function
	nlp['c'] = [0.]*len(nlp['c'])

	# change target reaction
	nlp['xl'][rxn_id[target_reaction]] = 0.
	nlp['xu'][rxn_id[target_reaction]] = 1000.
	nlp['c'][rxn_id[target_reaction]] = 1.

	return get_reduced_costs_from_nlp(nlp, objective_value)

def revert_gene_knockouts(model, genes):
	raise NotImplementedError

def single_gene_deletion(model, gene, threshold = 0.01, solver = 'qminos'):
	if solver not in [ 'gurobi', 'qminos' ]:
		raise Exception('The solver argument should be \'qminos\' or any valid for COBRApy models, such as \'gurobi\'.')
	if isinstance(gene, (list, set)):
		raise Exception('The method is limited to one gene only. Use model.perform_gene_knockouts(), followed by model.optimize() or model.feasibility().')

	test = perform_gene_knockouts(model, gene)

	if isinstance(model, coralme.core.model.MEModel):
		if test.feasibility({ test.mu : threshold }):
			return gene, False # gene is not essential
		else:
			return gene, True # gene is essential
	else:
		if solver == 'qminos':
			# feasibility developed to work with ME-models only
			# output is True or False; if True, test.solution is created
			coralme.core.model.MEModel.optimize(test)
		elif solver in ['gurobi']:
			test.solver = solver
			test.solution = test.optimize()
		else:
			test.solver = 'gurobi'

	# if sol.status != 'optimal' or sol.objective_value < threshold:
	if hasattr(test, 'solution'):
		if test.solution.status == 'infeasible':
			return gene, True # gene is essential
		elif test.solution.status == 'optimal' and test.solution.objective_value < threshold:
			return gene, True # gene is essential
		else:
			return gene, False # gene is not essential (over the threshold)
	else:
		return gene, True # gene is essential

if __name__ == '__main__':
	single_gene_deletion(model, gene, threshold = 0.01, solver = 'qminos')



import os
import sys
import copy
import json
import sympy
import jsonschema
from collections import OrderedDict

import logging
log_format = '%(asctime)s %(message)s' #%(clientip)-15s %(user)-8s

import pandas
import cobra
import coralme

try:
	# If cannot import SymbolicParameter, assume using cobrapy versions <= 0.5.11
	from optlang.interface import SymbolicParameter
except ImportError:
	from cobra.io.json import save_json_model
	from cobra.io.dict import _metabolite_from_dict
else:
	from cobra.io.json import _metabolite_from_dict as metabolite_from_dict
	from cobra.io.json import save_json_model

cur_dir = os.path.dirname(os.path.abspath(__file__))

def get_schema():
	"""
	Load JSON schema for ME-model JSON saving/loading

	Returns
	-------
	dict
		JSONSCHEMA

	"""
	with open(os.path.join(cur_dir, 'JSONSCHEMA'), 'r') as f:
		return json.load(f)

def save_json_me_model(model, file_name, sort = True, compress = False):
	"""
	Save a full JSON version of the ME-model. Saving/loading a model in this
	format can then be loaded to return a ME-model identical to the one saved,
	which retains all ME-model functionality.

	Parameters
	----------

	model : :class:`coralme.core.model.MEModel`
		A full ME-model

	file_name : str or file-like object
		Filename of the JSON output or an open json file

	"""

	#should_close = False
	#if isinstance(file_name, str):
		#file_name = open(file_name, 'w')
		#should_close = True

	# set logger
	log = logging.getLogger() # root logger
	for hdlr in log.handlers[:]: # remove all old handlers
		log.removeHandler(hdlr)

	# Old code works in a separate script; but it works if we remove the old handler
	logging.basicConfig(level = logging.WARNING, format = log_format)
	#log.addHandler(logging.StreamHandler(sys.stdout))
	logging.captureWarnings(True)

	model_dict = coralme.io.dict.me_model_to_dict(model)

	# cannot be before me_model_from_dict
	if 'default_parameters' in model.global_info:
		model_dict['global_info']['default_parameters'] = { str(k):v for k,v in model_dict['global_info']['default_parameters'].items() }

	logging.shutdown()

	# Confirm that dictionary representation of model adheres to JSONSCHEMA
	try:
		jsonschema.validate(model_dict, get_schema())
	except jsonschema.ValidationError:
		raise Exception('Must pass valid ME-model json file')

	#json.dump(model_dict, file_name)

	#if should_close:
		#file_name.close()

	def set_default(obj):
		if isinstance(obj, set):
			return list(obj)
		if isinstance(obj, pandas.DataFrame):
			return obj.to_dict()
		#raise TypeError

	json_str = json.dumps(model_dict, indent = 2, sort_keys = sort, default = set_default)
	if compress:
		import gzip
		with gzip.open(file_name + '.gz', 'w') as outfile:
			outfile.write(json_str.encode('utf-8'))
	else:
		with open(file_name, 'w') as outfile:
			outfile.write(json_str)

def load_json_me_model(file_name):
	"""
	Load a full JSON version of the ME-model. Loading a model in this format
	will return a ME-model identical to the one saved, which retains all
	ME-model functionality.

	Parameters
	----------
	file_name : str or file-like object
		Filename of the JSON output or an open json file

	Returns
	-------
	:class:`coralme.core.model.MEModel`
		A full ME-model

	"""
	# set logger
	log = logging.getLogger() # root logger
	for hdlr in log.handlers[:]: # remove all old handlers
		log.removeHandler(hdlr)

	# Old code works in a separate script; but it works if we remove the old handler
	logging.basicConfig(filename = 'MELoader.log', filemode = 'w', level = logging.WARNING, format = log_format)
	log.addHandler(coralme.builder.main.ListHandler([]))
	#log.addHandler(logging.StreamHandler(sys.stdout))
	logging.captureWarnings(True)

	if isinstance(file_name, str):
		with open(file_name, 'r') as f:
			model_dict = json.load(f)
	else:
		model_dict = json.load(file_name)

	try:
		jsonschema.validate(model_dict, get_schema())
	except jsonschema.ValidationError:
		raise Exception('Must pass valid ME-model json file')

	model = coralme.io.dict.me_model_from_dict(model_dict)

	logging.shutdown()

	# We will remove duplicates entries in the log output
	with open('MELoader.log', 'w') as outfile:
		logger = log.handlers[1].log_list

		tmp = pandas.DataFrame(logger)
		for idx, data in tmp.drop_duplicates(subset = 1).iterrows():
			outfile.write('{:s} {:s}\n'.format(data[0], data[1]))

	return model

# -----------------------------------------------------------------------------
# Functions below here facilitate json dumping/loading of reduced ME-models
# without all process_data/reaction info intact.

def save_reduced_json_me_model(me0, file_name):
	"""
	Save a stripped-down JSON version of the ME-model. This will exclude all of
	ME-model information except the reaction stoichiometry information and the
	reaction bounds. Saving/loading a model in this format will thus occur much
	quicker, but limit the ability to edit the model and use most of its
	features.

	Parameters
	----------
	me0 : :class:`coralme.core.model.MEModel`
		A full ME-model

	file_name : str or file-like object
		Filename of the JSON output

	"""
	me = copy.deepcopy(me0)

	for rxn in me.reactions:
		for met in rxn.metabolites:
			s = rxn._metabolites[met]
			if isinstance(s, sympy.Basic):
				rxn._metabolites[met] = str(s)
		if isinstance(rxn.lower_bound, sympy.Basic):
			rxn.lower_bound = str(rxn.lower_bound)
		if isinstance(rxn.upper_bound, sympy.Basic):
			rxn.upper_bound = str(rxn.upper_bound)

	for met in me.metabolites:
		if isinstance(met._bound, sympy.Basic):
			met._bound = str(met._bound)

	save_json_model(me, file_name)

def load_reduced_json_me_model(file_name):
	"""
	Load a stripped-down JSON version of the ME-model. This will exclude all of
	ME-model information except the reaction stoichiometry information and the
	reaction bounds. Saving/loading a model in this format will thus occur much
	quicker, but limit the ability to edit the model and use most of its
	features.

	Parameters
	----------
	file_name : str or file-like object
		Filename of the JSON ME-model

	Returns
	-------
	:class:`cobra.core.model.Model`
		COBRA Model representation of the ME-model. This will not include
		all of the functionality of a :class:`~coralme.core.model.MEModel` but
		will solve identically compared to the full model.
	"""
	if isinstance(file_name, str):
		with open(file_name, 'r') as f:
			obj = json.load(f)
	else:
		obj = json.load(file_name)

	model = cobra.Model()

	# If cannot import SymbolicParameter, assume using cobrapy
	# versions <= 0.5.11. If versions >= 0.8.0 are used, a ME-model interface
	# must be assigned as the solver interface
	try:
		from optlang.interface import SymbolicParameter
	except ImportError:
		pass
	else:
		model.solver = coralme.util.me_model_interface

	default_reactions = [i.id for i in model.reactions]

	for k, v in obj.items():
		if k in {'id', 'name'}:
			setattr(model, k, v)

	def _reaction_from_dict(reaction, model):
		new_reaction = cobra.Reaction()
		for k, v in reaction.items():
			if k in {'objective_coefficient', 'reversibility', 'reaction'}:
				continue
			elif k == 'metabolites':
				new_reaction.add_metabolites(OrderedDict(
					(model.metabolites.get_by_id(str(met)),
					 coralme.io.dict.get_numeric_from_string(coeff))
					for met, coeff in v.items()))
			elif k in {'upper_bound', 'lower_bound'}:
				v = coralme.io.dict.get_numeric_from_string(v)
				setattr(new_reaction, k, v)
			else:
				setattr(new_reaction, k, v)
		return new_reaction

	model.add_metabolites([metabolite_from_dict(metabolite) for metabolite in obj['metabolites']])

	new_reactions = [_reaction_from_dict(reaction, model) for reaction in obj['reactions']]

	model.remove_reactions(default_reactions)
	model.add_reactions(new_reactions)

	return model



#!/usr/bin/python3

# ME model troubleshooting
# Originally developed by JDTB@UCSD, 2022
# Modified by RSP@UCSD, 2022

import coralme
import cobra
import sympy
import pandas
import logging
import sys
log = logging.getLogger(__name__)

def process_model(model, growth_key = sympy.Symbol('mu', positive = True)):
	"""
	Get a dictionary containing information on whether a metabolite has
	producing or consuming reactions. This is used to find gaps.
	"""
	dct = {}
	for met in model.metabolites:
		filter1 = type(met) == cobra.core.metabolite.Metabolite or type(met) == coralme.core.component.Metabolite
		filter2 = met.id.startswith('trna')
		filter3 = met.id.endswith('trna_c')

		if filter1 and not filter2 and not filter3:
			t = { 'c' : set(), 'p' : set() }
			#seen = [] #?
			for rxn in met.reactions:
				if rxn.id.startswith('BIOMASS_'):
					continue

				lb, ub = rxn.lower_bound, rxn.upper_bound

				# Replace 'growth_key' if model is a ME-model
				if hasattr(lb, 'subs'):
					lb = lb.subs(growth_key, 1.)
				if hasattr(ub, 'subs'):
					ub = ub.subs(growth_key, 1.)

				coeff = rxn.metabolites[met]
				if hasattr(coeff, 'subs'):
					coeff = coeff.subs(growth_key, 1.)

				pos = 1 if coeff > 0 else -1
				rev = 1 if lb < 0 else 0
				fwd = 1 if ub > 0 else 0
				if pos*fwd == -1 or pos*rev == +1:
					t['c'].add(rxn.id)
				if pos*fwd == +1 or pos*rev == -1:
					t['p'].add(rxn.id)
			dct[met.id] = t
	return dct

def add_exchange_reactions(me, metabolites, prefix = 'SK_'):
	"""Add exchange/sink reaction to the model"""
	rxns = []
	for met in metabolites:
		rxn_id = prefix + met
		if rxn_id not in me.reactions:
			r = coralme.core.reaction.MEReaction(rxn_id)
			me.add_reactions([r])
			r.add_metabolites({ met: -1 })
		else:
			r = me.reactions.get_by_id(rxn_id)
		r.bounds = (-10, 1000)
		rxns.append(r)
		#print(r.id,r.lower_bound,r.upper_bound,r.reaction)
	return rxns

def find_gaps(model, growth_key = sympy.Symbol('mu', positive = True)):
	"""Find gaps in the model"""
	g = {}
	dct = process_model(model, growth_key = growth_key)
	for met, t in dct.items():
		# not producing, not consuming, not uerever
		g[met] = { 'p' : 0, 'c' : 0, 'u' : 0 }
		if not t['c']:
			g[met]['c'] = 1
		if not t['p']:
			g[met]['p'] = 1
		if len(t['c']) == 1 and t['c'] == t['p']:
			g[met]['u'] = 1
	df = pandas.DataFrame.from_dict(g).T
	df = df[df.any(axis = 1)]
	df = df.sort_index()
	return df

def find_issue(query,d,msg = ''):
	"""Retrieve any warning message in curation notes associated with an ID"""
	if isinstance(d,dict):
		if 'msg' in d:
			msg = d['msg']
			if 'triggered_by' in d:
				trigger = d['triggered_by']
				find_issue(query,trigger,msg=msg)
		else:
			for k,v in d.items():
				find_issue(query,v,msg=msg)
	elif isinstance(d,list):
		for i in d:
			find_issue(query,i,msg=msg)
	elif isinstance(d,str):
		if query == d:
			print(msg)
	else:
		raise TypeError("unsupported type  " + type(d))

def fill_builder(b,fill_with='CPLX_dummy',key=None,d=None,fieldname=None,warnings=None):
	"""Fill empty fields in builder with CPLX dummy"""
	if isinstance(b,coralme.builder.main.MEBuilder):
		for i in dir(b.org):
			if i[0] == '_':
				continue
			attr = getattr(b.org,i)
			if not isinstance(attr,dict):
				continue
			fill_builder(attr,fill_with=fill_with,fieldname=i,warnings = warnings)
	elif isinstance(b,dict):
		for k,v in b.items():
			fill_builder(v,key=k,d=b,fill_with=fill_with,fieldname=fieldname,warnings=warnings)
	elif isinstance(b,list):
		include_keys = ['enzymes','proteins','enzyme','protein','machine']
		for ik in include_keys:
			if key in ik:
				if not b:
					d[key] = ['CPLX_dummy']
	elif isinstance(b,str):
		include_keys = ['enzymes','proteins','enzyme','protein','machine']
		for ik in include_keys:
			if key in ik:
				if not b:
					d[key] = 'CPLX_dummy'
	else:
		pass

def gap_find(me_model,de_type = None):
	"""Find and classify gaps in the model"""

	logging.warning('  '*5 + 'Finding gaps in the ME-model...')
	me_gaps = coralme.builder.troubleshooting.find_gaps(me_model, growth_key = me_model.mu)

	if de_type == 'me_only':
		logging.warning('  '*5 + 'Finding gaps from the M-model only...')
		m_gaps = coralme.builder.troubleshooting.find_gaps(me_model.gem)
		idx = list(set(me_gaps.index) - set(m_gaps.index))
	else:
		idx = list(set(me_gaps.index))
	new_gaps = me_gaps.loc[idx]

	filt1 = new_gaps['p'] == 1
	filt2 = new_gaps['c'] == 1
	filt3 = new_gaps['u'] == 1

	deadends = list(new_gaps[filt1 | filt2 | filt3].index)
	deadends = sorted([ x for x in deadends if 'biomass' not in x if not x.endswith('_e') ])

	logging.warning('  '*5 + '{:d} metabolites were identified as deadends.'.format(len(deadends)))
	for met in deadends:
		name = me_model.metabolites.get_by_id(met).name
		logging.warning('  '*6 + '{:s}: {:s}'.format(met, 'Missing metabolite in the M-model.' if name == '' else name))
	return deadends

def gap_fill(me_model, deadends = [], growth_key_and_value = { sympy.Symbol('mu', positive = True) : 0.1 }, met_types = 'Metabolite',solver="qminos"):
	"""Add sink reactions of gap metabolites to the model"""
	if solver in ['gurobi', 'cplex']:
		me_model.get_solution = me_model.optimize_windows
		me_model.get_feasibility = me_model.feas_windows(solver = solver)
	elif solver == "qminos":
		me_model.get_solution = me_model.optimize
		me_model.get_feasibility = me_model.feasibility
	# if sys.platform == 'win32':
	# 	me_model.get_solution = me_model.opt_gurobi
	# 	me_model.get_feasibility = me_model.feas_gurobi
	# else:
	# 	me_model.get_solution = me_model.optimize
	# 	me_model.get_feasibility = me_model.feasibility

	if len(deadends) != 0:
		logging.warning('  '*5 + 'Adding a sink reaction for each identified deadend metabolite...')
		coralme.builder.troubleshooting.add_exchange_reactions(me_model, deadends, prefix='TS_')
	else:
		logging.warning('  '*5 + 'Empty set of deadends metabolites to test.')
		return None

	logging.warning('  '*5 + 'Optimizing gapfilled ME-model...')

	if me_model.get_feasibility(keys = growth_key_and_value):
		#logging.warning('  '*5 + 'The ME-model is feasible.')
		logging.warning('  '*5 + 'Gapfilled ME-model is feasible with growth rate {:g} 1/h.'.format(list(growth_key_and_value.values())[0]))
		return True
	else:
		#logging.warning('  '*5 + 'The ME-model is not feasible.')
		logging.warning('  '*5 + 'Provided set of sink reactions for deadend metabolites does not allow growth.')
		return False

def brute_force_check(me_model, metabolites_to_add, growth_key_and_value,solver="qminos"):
	"""
	Iteratively search for minimal set of metabolites that are needed as
	sinks to allow for growth. This function searches by batches of
	different types of metabolites.
	"""
	if solver in ['gurobi', 'cplex']:
		me_model.get_solution = me_model.optimize_windows
		me_model.get_feasibility =me_model.feas_windows(solver = solver)
	elif solver == "qminos":
		me_model.get_solution = me_model.optimize
		me_model.get_feasibility = me_model.feasibility
	# if sys.platform == 'win32':
	# 	me_model.get_solution = me_model.opt_gurobi
	# 	me_model.get_feasibility = me_model.feas_gurobi
	# else:
	# 	me_model.get_solution = me_model.optimize
	# 	me_model.get_feasibility = me_model.feasibility

	logging.warning('  '*5 + 'Adding sink reactions for {:d} metabolites...'.format(len(metabolites_to_add)))
# 	existing_sinks = [r.id for r in me_model.reactions.query('^TS_')]
	sk_rxns = coralme.builder.troubleshooting.add_exchange_reactions(me_model, metabolites_to_add, prefix='TS_')

	if me_model.get_feasibility(keys = growth_key_and_value):
		pass
	else:
		logging.warning('  '*5 + 'Provided metabolites through sink reactions cannot recover growth. Proceeding to next set of metabolites.')
		return metabolites_to_add, [], False

	rxns = []
	rxns_to_drop = []
# 	rxns_to_append = []
# 	for idx, flux in me_model.solution.fluxes.items():
	for r in sk_rxns:
		idx = r.id
		flux = me_model.solution.fluxes[idx]
		if idx.startswith('TS_') and idx.split('TS_')[1] in metabolites_to_add:
# 			if r.id in existing_sinks:
# 				rxns_to_append.append(idx)
# 				continue
			if abs(flux) > 0:
				rxns.append(idx)
			else:
				#logging.warning('Closing {}'.format(idx))
				rxns_to_drop.append(idx)
				me_model.reactions.get_by_id(idx).bounds = (0, 0)

	logging.warning('  '*6 + 'Sink reactions shortlisted to {:d} metabolites.'.format(len(rxns)))

	# reaction_id:position in the model.reactions DictList object
# 	rxns = rxns + rxns_to_append# Try present SKs the last.
# 	logging.warning('  '*6 + 'Will try a total of {:d} metabolites including previous iterations:'.format(len(rxns)))
	ridx = []
	for r in rxns:
		ridx.append((r,me_model.reactions._dict[r]))
# 	ridx = { k:v for k,v in me_model.reactions._dict.items() if k in rxns }

	# populate with stoichiometry
	Sf, Se, lb, ub, b, c, cs, atoms, lambdas, Lr, Lm = me_model.construct_lp_problem()

	if lambdas is None:
		Sf, Se, lb, ub = coralme.builder.helper_functions.evaluate_lp_problem(Sf, Se, lb, ub, growth_key_and_value, atoms)
	else:
		Sf, Se, lb, ub = coralme.builder.helper_functions.evaluate_lp_problem(Sf, lambdas, lb, ub, growth_key_and_value, atoms)

	res = []
	msg = 'Processed: {:s}/{:d}, Gaps: {:d}. The ME-model is {:s}feasible if {:s} is closed.'
	for idx, (rxn, pos) in enumerate(ridx):
		lb[pos] = 0
		ub[pos] = 0
		if me_model.get_feasibility(keys = growth_key_and_value, **{'lp' : [Sf, dict(), lb, ub, b, c, cs, set(), lambdas, Lr, Lm]}):
			res.append(False)
			logging.warning('{:s} {:s}'.format('  '*6, msg.format(str(idx+1).rjust(len(str(len(ridx)))), len(ridx), len([ x for x in res if x ]), '', rxn)))
		else:
			lb[pos] = -1000
			ub[pos] = +1000
			res.append(True)
			logging.warning('{:s} {:s}'.format('  '*6, msg.format(str(idx+1).rjust(len(str(len(ridx)))), len(ridx), len([ x for x in res if x ]), 'not ', rxn)))

	bf_gaps = [ y for x,y in zip(res, rxns) if x ] # True
	no_gaps = [ y for x,y in zip(res, rxns) if not x ] + rxns_to_drop

	return bf_gaps, no_gaps, True

def get_mets_from_type(me_model,met_type):
	"""Get metabolites by type relevant to gap filling the model"""
	if met_type[1] == 'User guesses':
		return set(met_type[0])
	elif met_type == 'ME-Deadends':
		return set(coralme.builder.troubleshooting.gap_find(me_model,de_type='me_only'))
	elif met_type == 'All-Deadends':
		return set(coralme.builder.troubleshooting.gap_find(me_model))
	elif met_type == 'Cofactors':
		return set(get_cofactors_in_me_model(me_model))
	else:
		mets = set()
		for met in me_model.metabolites:
			filter1 = type(met) == getattr(coralme.core.component, met_type)
			filter2 = met.id.startswith('trna')
			filter3 = met.id.endswith('trna_c')
			filter4 = met.id.endswith('_e')
			if filter1 and not filter2 and not filter3 and not filter4:
				mets.add(met.id)
		return mets

def _append_metabolites(mets,new_mets):
	"""Merge metabolite lists"""
	return mets + [m for m in new_mets if m not in mets]

def brute_check(me_model, growth_key_and_value, met_type, skip = set(), history = dict(),solver="qminos"):
	"""Remove metabolites from our heuristics and call the brute force search algorithm"""
	mets = get_mets_from_type(me_model,met_type)
	if met_type == 'Metabolite':
		#remove from the metabolites to test that are fed into the model through transport reactions
		medium = set([ '{:s}_c'.format(x[3:-2]) for x in me_model.gem.medium.keys() ])
		mets = set(mets).difference(medium)
		# filter out manually
		mets = set(mets).difference(set(['ppi_c', 'ACP_c', 'h_c']))
		mets = set(mets).difference(set(['adp_c', 'amp_c', 'atp_c']))
		mets = set(mets).difference(set(['cdp_c', 'cmp_c', 'ctp_c']))
		mets = set(mets).difference(set(['gdp_c', 'gmp_c', 'gtp_c']))
		mets = set(mets).difference(set(['udp_c', 'ump_c', 'utp_c']))
		mets = set(mets).difference(set(['dadp_c', 'dcdp_c', 'dgdp_c', 'dtdp_c', 'dudp_c']))
		mets = set(mets).difference(set(['damp_c', 'dcmp_c', 'dgmp_c', 'dtmp_c', 'dump_c']))
		mets = set(mets).difference(set(['datp_c', 'dctp_c', 'dgtp_c', 'dttp_c', 'dutp_c']))
		mets = set(mets).difference(set(['nad_c', 'nadh_c', 'nadp_c', 'nadph_c']))
		mets = set(mets).difference(set(['5fthf_c', '10fthf_c', '5mthf_c', 'dhf_c', 'methf_c', 'mlthf_c', 'thf_c']))
		mets = set(mets).difference(set(['fad_c', 'fadh2_c', 'fmn_c']))
		mets = set(mets).difference(set(['coa_c']))
	mets = set(mets).difference(skip)
	if met_type[1] == 'User guesses':
		history['User guesses'] = mets
	else:
		history[met_type] = mets

	mets_to_check = []
	for k,v in history.items():
		mets_to_check = _append_metabolites(mets_to_check,v)
	return history,coralme.builder.troubleshooting.brute_force_check(me_model,
															  mets_to_check[::-1],
															  growth_key_and_value,
																	solver=solver)

def get_cofactors_in_me_model(me):
	"""Get metabolites that work as cofactors in the model"""
	cofactors = set()
	for i in me.process_data.query('^mod_'):
		for k,v in i.stoichiometry.items():
			if not me.metabolites.has_id(k):
				continue
			if v < 0:
				cofactors.add(k)
	return list(cofactors)



# -*- coding: utf-8 -*-

# Always prefer setuptools over distutils
#from ez_setup import use_setuptools
#use_setuptools()
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path, walk
import versioneer

def main():
	# additional files
	data_files = []
	for dirpath, dirnames, filenames in walk('coralme/iJL1678b-ME'):
		tmp = []
		for filename in filenames:
			tmp.append(path.join(dirpath, filename))
		data_files.append(('coralme', tmp))

	# Get the long description from the README file
	here = path.abspath(path.dirname(__file__))
	with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
		long_description = f.read()

	setup(
		name='coralME',
		license='MIT',
		#version='1.0', # use tags

		version=versioneer.get_version(), # from versioneer
		cmdclass=versioneer.get_cmdclass(), # from versioneer

		description='Comprehensive Reconstruction Algorithm for ME-models (coralME)',
		long_description=long_description,
		long_description_content_type='text/markdown',
		url='https://github.com/jdtibochab/coralme',
		author='Juan D. Tibocha-Bonilla and Rodrigo Santibanez-Palominos',
		author_email='jdtibochab@users.noreply.github.com',
		classifiers=[
			#'Development Status :: 1 - Planning',
			#'Development Status :: 2 - Pre-Alpha',
			#'Development Status :: 3 - Alpha',
			#'Development Status :: 4 - Beta',
			'Development Status :: 5 - Production/Stable',
			#'Development Status :: 6 - Mature',
			#'Development Status :: 7 - Inactive',

			# Indicate who your project is intended for
			'Intended Audience :: Science/Research',
			'Topic :: Scientific/Engineering :: Bio-Informatics',

			# Pick your license as you wish
			'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

			# Specify the Python versions you support here. In particular, ensure
			# that you indicate whether you support Python 2, Python 3 or both.
			#'Programming Language :: Python :: 2',
			#'Programming Language :: Python :: 2.7',
			'Programming Language :: Python :: 3',
			#'Programming Language :: Python :: 3.4',
			#'Programming Language :: Python :: 3.5',
			#'Programming Language :: Python :: 3.6',
			'Programming Language :: Python :: 3.7',
			'Programming Language :: Python :: 3.8',
			'Programming Language :: Python :: 3.9',
			'Programming Language :: Python :: 3.10',

			# other classifiers added by author
			'Environment :: Console',
			'Operating System :: Unix',
			'Operating System :: Microsoft :: Windows',
			'Operating System :: MacOS',
		],

		python_requires='~=3.0',
		keywords=[
			'metabolism',
			'biology',
			'constraint-based',
			'linear programming',
			'mixed-integer',
			'optimization',
			'flux-balance analysis',
			'reconstruction'
			],
		install_requires=[
			'cobra<=0.29.0',
			'python-libsbml<=5.20.1',
			'Biopython==1.80',
			'anyconfig<=0.13.0',
			'pyranges<=0.0.129',
			'XlsxWriter<=3.1.2',
			'openpyxl<=3.1.2',
			'numpy<=1.26.4',
			'scipy<=1.11.1',
			'sympy<=1.12',
			'pandas<=1.5.1',
			'jsonschema<=4.21.1',
			'tqdm<=4.62.3',
			],

		# WARNING: seems to be bdist_wheel only
		packages=find_packages(exclude=('contrib', 'docs', 'tests', 'tutorials')),
		# using the MANIFEST.in file to exclude same folders from sdist
		include_package_data=False,

		# WARNING: use this way to install in the package folder and
		# to have access to files using importlib_resources or importlib.resources
		# bdist_wheel only
		package_data = {
			'coralme' : [
				'iJL1678b-ME/*',
				'iJL1678b-ME/building_data/*',
				'solver/*.so',
				'io/JSONSCHEMA',
				'builder/column_format.json',
				'licenses/*'
				]
			},

		# WARNING: do not use data_files as the installation path is hard to determine
		# e.g.: ubuntu 18.04 installs to /usr/local/installation_path or /$USER/.local/installation_path
		#data_files=[('installation_path', ['filename'])],
		#data_files=data_files,

		# others
		project_urls={
			'Manual': 'https://coralme.readthedocs.io',
			'Bug Reports': 'https://github.com/jdtibochab/coralme/issues',
			'Source': 'https://github.com/jdtibochab/coralme',
		},
	)

if __name__ == '__main__':
    main()



import pickle

def load_pickle_me_model(path):
    with open(path, "rb") as infile:
        return pickle.load(infile)

def save_pickle_me_model(me, path):
    with open(path, "wb") as outfile:
        pickle.dump(me, outfile)



import numpy
import pandas
import cobra

def ToExcel(model, outfile: str):
	#metabolites
	try:
		mets = {
			'notes' : [],
			'model' : [],
			'_id' : [],
			'name' : [],
			'formula' : [],
			'compartment' : [],
			'charge' : [],
			'_annotation' : [],
			}

		for metabolite in model.metabolites:
			for key in list(mets.keys()):
				if key in ['model', 'notes']:
					mets[key].append(numpy.nan)
				else:
					mets[key].append(metabolite.__dict__[key])

		mets = pandas.DataFrame.from_dict(mets)
		# formating annotations
		tmp = pandas.json_normalize(mets['_annotation'])
		tmp = tmp.astype(str)
		tmp = tmp.apply(lambda col: col.str.replace('[\'', '', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('\']', '', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('\', \'', ';', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('nan', '', regex = False))
		# add annotations
		mets = pandas.concat([mets.iloc[:, :-1], tmp], axis = 1)

	except:
		print('metabolite dump failed')

	#reactions
	try:
		rxns = {
			'notes' : [],
			'model' : [],
			'_id' : [],
			'name' : [],
			'_metabolites' : [],
			'_lower_bound' : [],
			'_upper_bound' : [],
			'_gpr' : [],
			'subsystem' : [],
			'_annotation' : [],
			}

		for reaction in model.reactions:
			rxn_subs = [ [x,y] for x,y in zip([x._id for x in reaction._metabolites], reaction._metabolites.values()) if y < 0]
			rxn_prod = [ [x,y] for x,y in zip([x._id for x in reaction._metabolites], reaction._metabolites.values()) if y > 0]

			for key in rxns.keys():
				if key in ['model']:
					if reaction.objective_coefficient != 0.:
						rxns[key].append('__OBJ__')
					else:
						rxns[key].append(numpy.nan)

				elif key in ['notes']:
					if reaction.objective_coefficient != 0.:
						rxns[key].append('__BIOMASS__')
					else:
						rxns[key].append(numpy.nan)

				elif key == '_metabolites':
					rxns['_metabolites'].append(
						'{:s} = {:s}'.format(
							' + '.join([ '{:.6g} {:s}'.format(y*-1., x) for x,y in rxn_subs ]),
							' + '.join([ '{:.6g} {:s}'.format(y*+1., x) for x,y in rxn_prod ])))

				elif key == '_gpr':
					rxns['_gpr'].append(reaction._gpr.to_string())

				else:
					rxns[key].append(reaction.__dict__[key])

		rxns = pandas.DataFrame.from_dict(rxns)
		# formating annotations
		tmp = pandas.json_normalize(rxns['_annotation'])
		tmp = tmp.astype(str)
		tmp = tmp.apply(lambda col: col.str.replace('[\'', '', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('\']', '', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('\', \'', ';', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('nan', '', regex = False))
		# add annotations
		rxns = pandas.concat([rxns.iloc[:, :-1], tmp], axis = 1)

	except:
		print('reactions dump failed')

	# genes
	if True:
		genes = {
			'_id' : [],
			'name' : [],
			'_annotation' : [],
			}

		for gene in model.genes:
			for key in genes.keys():
				genes[key].append(gene.__dict__[key])

		genes = pandas.DataFrame.from_dict(genes)
		# formating annotations
		tmp = pandas.json_normalize(genes['_annotation'])
		tmp = tmp.astype(str)
		tmp = tmp.apply(lambda col: col.str.replace('[\'', '', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('\']', '', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('\', \'', ';', regex = False))
		tmp = tmp.apply(lambda col: col.str.replace('nan', '', regex = False))
		# add annotations
		genes = pandas.concat([genes.iloc[:, :-1], tmp], axis = 1)

	else:
		print('genes dump failed')

	if outfile.endswith('.xlsx'):
		with open(outfile, 'wb') as outfile:
			writer = pandas.ExcelWriter(outfile, engine = 'xlsxwriter')

			rxns.to_excel(writer, index = False, sheet_name = 'reactions')
			mets.to_excel(writer, index = False, sheet_name = 'metabolites')
			genes.to_excel(writer, index = False, sheet_name = 'genes')

			for data, sheet in zip([ rxns, mets, genes ], [ 'reactions', 'metabolites', 'genes' ]):
				(max_row, max_col) = data.shape

				# Get the xlsxwriter workbook and worksheet objects
				workbook  = writer.book
				worksheet = writer.sheets[sheet]

				# Freeze first row
				worksheet.freeze_panes(1, 0)

				# Set the autofilter
				worksheet.autofilter(0, 0, max_row, max_col - 1)

				# Make the columns wider for clarity
				worksheet.set_column_pixels(0,  max_col - 1, 96)

				# Set zoom level
				worksheet.set_zoom(120)

			# Close the Pandas Excel writer and output the Excel file.
			writer.close()

	return None
